( Recording Program for Programmable Digital Synthesizer running 3-instruction Forth on Atmel 328 Arduino Nano )

\ THIS NEEDS FAIR BIT OF WORD TO GET IT RECORDING PROPERLY

: record  ( on1/off0 -- )   \ turns on/off piano serial interface -- when a note is played, piano sends digital note data
    xstore CALL_RECORD xcall ; 
\ : [..] ( n -- idx ) \ plays note from scale array, 1 through 8 is from scale8
\ 	1 - scale + c@ ;

\ * words for PARSING digital piano note data * 
variable numnotes     \ number of entries in score
	: numnotes++ ( -- ) numnotes c@ 1 + numnotes c! ;
	: reset-notes ( -- ) 0 numnotes c! ;
	: numnotes@ numnotes c@ ;
    : del-note  numnotes c@ 1 - numnotes c! ;

\ * DATA STRUCTURE:  note duration rest, each is 2 bytes *
create notes 255 wwords allot  \ holds notes as idx into piano frequency array (later could be frequencies if needed)
create dur   255 wwords allot  \ holds durations
create rest  255 wwords allot  \ holds rest before indexed note
create score 3 dwords allot    \ holds pointers to note, duration, and rests
: init-score  ( -- ) \ notes dur rest
        rest dur notes score !+ !+ !+ drop 
        reset-notes ( 3 0 set-zero ) ;
\ fetches the individual pointers
: *n ( -- n-ptr ) score @ ;
: *d ( -- d-ptr ) score 1 dwords + @ ;
: *r ( -- r-ptr ) score 2 dwords + @ ;
\ save note data
: !n ( n -- )    *n  numnotes@  wwords +    ww! ;
: !d ( dur -- )  *d   numnotes@ wwords +    ww! ;
: !r ( rest -- ) *r   numnotes@ wwords +    ww! ;
\ show note data
: show-notes  cr numnotes@ dup if 0 do *n i wwords +  ww@ . ." .. "  loop cr else drop then ;
: show-dur  cr numnotes@ 0 do *d i wwords +  ww@ . ." " loop cr ;
: show-rest  cr numnotes@ 0 do *r i wwords +  ww@ . ." " loop cr ;

\ setup
: record-on  1 record   init-score   ( 3 0 set-zero ) ;
: record-off 0 record  ( reset-zero ) ;

: [parse-note] ( adr -- adr' val16 )  \ reads byte buffer parsed as little-endian words with low byte first to val16
    c@+ swap      c@+ 8 lshift     rot    or ;
: parse-note ( adr -- idx )  \ 13 fields to a note record
  rcv-buffer
  [parse-note] ( dup   hex. ) !d
  [parse-note] ( dup   hex. ) !r
  [parse-note] ( dup   hex. ) !n
  numnotes++
  drop ;

\ with the number decoded, play it using ., relative to 3 0 set-zero (or the start note of the piano key).  Or could remap to a scale and use ..

\ Simplified receive - key: it is not a realtime read

\ 6-byte key press tuples
: fetch-note-data  ( <serial> -- numbytes )  \ reads from serial buffer into rcv-buffer
    6 read-com ;
: show-note-data
    rcv-buffer 6 show-rcv 2drop ;
: extract-note ( -- note-id ) 
    rcv-buffer 2 wwords + [parse-note] nip ;


\ ============= STARTING TO INCORPORATE SOUND BOARD for arpeggios 

\ include 3iSound-4x4.fs
\ arpeggios - remove comment to have them repeat endlessly till key pressed

\ key?
\ key
\ include 3iRecord.fs

\ 1 record  \ need to start recording for this to work!
: press?  ( -- 0/1 ) fetch-note-data ;
extract-note

\ KEY ARPEGGIOS
 : c-arpeg  cc      mj  c_  note  begin   e_  note  g_  note  e_   note c_   note  key?  until key ;
 : d-arpeg  dd      mj  d_  note  begin   f#_ note  a_  note  f#_  note d_   note  key?  until key ; 
 : e-arpeg  ee      mj  e_  note  begin   g#_ note  b_  note  g#_  note e_   note  key?  until key ; 
 : f-arpeg  ff      mj  f_  note  begin   a_  note  c'  note  a_   note f_   note  key?  until key ;
 : g-arpeg  gg      mj  g_  note  begin   b_  note  d'  note  b_   note g_   note  key?  until key ;
 : dm-arpeg dd      mn  d_  note  begin   f_  note  a_  note  f_   note d_   note  key?  until key ;
 : a-arpeg  aa      mj  a_  note  begin   c#' note  e'  note  c#'  note a_   note  key?  until key ;
 : b-arpeg  bb      mj  b_  note  begin   d#' note  f#' note  d#'  note b_   note  key?  until key ;
 : b%-arpeg bb %%   mj  b%_ note  begin   d'  note  f'  note  d'   note b%_  note  key?  until key ;  \ 2 4 5  2 4 5  2 4 5 6# 6 4 5
 : c'-arpeg cc 7 +  mj  c'  note  begin   e'  note  g'  note  e'   note c'   note  key?  until key ;
 create key-arpeggios    ' c-arpeg , ' d-arpeg , ' e-arpeg , ' f-arpeg , ' g-arpeg , ' a-arpeg , ' b-arpeg , ' c'-arpeg , 
 : key-arpeg ( note-id -- ) 
    begin 1 - dwords key-arpeggios + @ execute 48 - dup 1 8 is_between? invert until drop ;

\ PRESS ARPEGGIOS
\ TURN RECORD ON FOR THIS
: key/press?   press? if extract-note -1 else key? if key drop 0 -1 else 0 then then ;
: c-arpegp  cc      mj      begin   e_  note  g_  note  e_   note  c_   note  key/press?  until ;
: d-arpegp  dd      mj      begin   f#_ note  a_  note  f#_  note  d_   note  key/press?  until ; 
: e-arpegp  ee      mj      begin   g#_ note  b_  note  g#_  note  e_   note  key/press?  until ; 
: f-arpegp  ff      mj      begin   a_  note  c'  note  a_   note  f_   note  key/press?  until ;
: g-arpegp  gg      mj      begin   b_  note  d'  note  b_   note  g_   note  key/press?  until ;
: dm-arpegp dd      mn      begin   f_  note  a_  note  f_   note  d_   note  key/press?  until ;
: a-arpegp  aa      mj      begin   c#' note  e'  note  c#'  note  a_   note  key/press?  until ;
: b-arpegp  bb      mj      begin   d#' note  f#' note  d#'  note  b_   note  key/press?  until ;
: b%-arpegp bb %%   mj      begin   d'  note  f'  note  d'   note  b%_  note  key/press?  until ;  \ 2 4 5   2 4 5   2 4 5 6# 6 4 5
: c'-arpegp cc 12 + mj      begin   e'  note  g'  note  e'   note  c'   note  key/press?  until ;
 create press-arpeggios  ' c-arpegp , ' d-arpegp , ' e-arpegp , ' f-arpegp , ' g-arpegp , ' a-arpegp , ' b-arpegp , ' c'-arpegp ,  
 : press-arpegp ( note-id -- ) 
    dup .. begin 1 - dwords press-arpeggios + @ execute dup 0 = if drop -1 else 0 then until ;
 
 
 \ revised to 1 5 3 5 picking pattern for sound of silence
 : [dm-arpeg] dd mn     d_   note   a_  note  f_  note  a_  note  ;
 : [c-arpeg]  cc mj     c_   note   g_  note  e_  note  g_  note  ;
 : [d-arpeg]  dd mj     d_   note   a_  note  f#' note  a_  note  ; 
 : [f-arpeg]  ff mj     f_   note   c'  note  a_  note  c'  note  ;
 : [g-arpeg]  gg mj     g_   note   d'  note  b_  note  d'  note  ;
 : [a-arpeg]  aa mj     a_   note   e'  note  c#' note  e'  note  ;
 : [b%-arpeg] bb %% mj  b%_  note   f'  note  d'  note  f'  note ;
 
: // 200 ms ;
: 1-2-3-4  4 0 do 32-on // clear32 // loop ;
: end-song 32-on // clear32 2 0 do // 32-on // clear32 loop ;
 
 \ sound of silence
 : silence
   clear32
   1-2-3-4
   // [dm-arpeg]  [dm-arpeg]  [dm-arpeg]  [dm-arpeg]
   // [c-arpeg]   [c-arpeg]   [c-arpeg]   [c-arpeg]  
   // [f-arpeg]   [f-arpeg]   [b%-arpeg]  [f-arpeg]
   // [f-arpeg]   [f-arpeg]   [b%-arpeg]  [f-arpeg]
   // [f-arpeg]   [b%-arpeg]  [b%-arpeg]  [b%-arpeg] [f-arpeg] 
   // [dm-arpeg]  [dm-arpeg]  [f-arpeg]   [c-arpeg]  [c-arpeg]   
   // [dm-arpeg]  [dm-arpeg]  [dm-arpeg]  d_ note
   end-song
 ;
 
\ include 3iRecord.fs

: arpeggiator 
  begin   fetch-note-data 0 > if  extract-note then 
  until ;

\ =======================================




: get-note ( -- note/stop )  \ read note from serial, 0 means no note, 1-13 notes
    numnotes@ $FF < if fetch-note-data dup if \ show-rcv cr
			  parse-note then 
    else cr ." Memory full.  reset-notes or del-note." cr
	then ;
: get-notes ( N -- ) 0 cr do get-note drop loop ;
: fix-first-note-rest  ( numnotes-first -- ) wwords *r + 5000 swap ww! ;

: all-notes ( -- )  cr numnotes@ >r begin get-note 0= until cr numnotes@ . ." notes logged." cr r> fix-first-note-rest show-notes ;
\ : play-notes-only ( -- ) 
\     numnotes@  dup  if 0 do 
\ 		*n i wwords + ww@    ., 
\ 	loop else drop then ;
\ : play-notes-fixedr ( -- ) 
\     numnotes@  dup  if 0 do 
\ 		R2   \ fixed rest
\ 		*n i wwords + ww@    1 - show-zero + freq lohi 
\ 		*d i wwords + ww@ 30 * lohi
\ 		speaker 
\ 		[tone] 
\ 	loop else drop then ;
: spin ( n -- ) 0 do loop ;
\ : play-notes-fixedd ( -- ) 
\     numnotes@  dup  if 0 do 
\ 		*r i wwords + ww@ 1000 / 1 + 100 * ms 
\ 		*n i wwords + ww@ 1 - show-zero + freq lohi 
\ 		100 lohi  \ fixed duration
\ 		speaker 
\ 		[tone]
\ 	loop else drop then ;
\ : play-notes ( -- ) 
\     numnotes@  dup  if 0 do 
\ 		*r i wwords + ww@ 1000 / 1 + 100 * ms 
\ 		*n i wwords + ww@ 1 - show-zero + freq lohi 
\ 		*d i wwords + ww@ 30 * lohi
\ 		speaker 
\ 		[tone] 
\ 	loop else drop then ;
\ : scale-notes ( -- ) 
\     numnotes@  dup  if 0 do 
\ 		*r i wwords + ww@ 1000 / 1 + 80 * ms 
\ 		*n i wwords + ww@ [..] freq lohi 
\ 		*d i wwords + ww@ 30 * lohi
\ 		speaker 
\ 		[tone]
\ 	loop else drop then ;
\ : scale-play ( -- ) 
\     cr numnotes@  dup  
\ 	if 0 
\ 		do 
\ 			R1
\ 			*n i wwords + ww@ dup >r [..] freq lohi 
\ 			*d i wwords + ww@ 30 * lohi
\ 			speaker
\ 			[tone] 
\ 			key r> over '-' = if  drop else . ." .. " 
\ 			dup '1' = if ." R1 " then '2' = if ." R2 " then then
\ 		loop 
\ 	else drop 
\ 	then ;

: help-record
  cr ." To record on your digital piano: "
  cr ."    record-on"
  cr ."    <play notes>"
  cr ."    get-note   all-notes  play-notes   del-note  reset-notes"
  cr ."    show-notes  show-dur  show-rest"
  cr ."    record-off"
;


\ =====================
\ V1 - 
\ V2 - 
\ V3 - 2021-05-15---Sat---16:57 - Integrated with Arpeggitor on piano board with lights board and keyboard (not 4x4"

